from .base import Widget
from toga.constants import *


class Label(Widget):
    def __init__(self, text, id=None, style=None, alignment=LEFT_ALIGNED):
        super().__init__(id=id, style=style, text=text, alignment=alignment)

    def _configure(self, text, alignment):
        self.text = text
        self.alignment = alignment
        
    @property
    def alignment(self):
        return self._alignment

    @alignment.setter
    def alignment(self, value):
        self._alignment = value
        self._set_alignment(value)

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        if value is None:
            self._text = ''
        else:
            self._text = str(value)
        self._set_text(value)
        self.rehint()

    def _set_alignment(self, value):
        raise NotImplementedError('Label widget must define _set_alignment()')

    def _set_text(self, value):
        raise NotImplementedError('Label widget must define _set_text()')



from .base import Widget


class WebView(Widget):
    def __init__(self, id=None, style=None, url=None, on_key_down=None):
        super(WebView, self).__init__(id=id, style=style, url=url, on_key_down=on_key_down)

    def _configure(self, url, on_key_down):
        self.url = url
        self.on_key_down = on_key_down

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value
        self._set_url(value)

    def set_content(self, root_url, content):
        self._url = root_url
        self._set_content(root_url, content)

    def _set_url(self, value):
        raise NotImplementedError('Webview widget must define _set_url()')

    def _set_content(self, root_url, content):
        raise NotImplementedError('Webview widget must define _set_content()')

    def evaluate(self, javascript):
        raise NotImplementedError('Webview widget must define evaluate()')



from .base import Widget


class Button(Widget):
    def __init__(self, label, id=None, style=None, on_press=None):
        super().__init__(id=id, style=style, label=label, on_press=on_press)

    def _configure(self, label, on_press):
        self.label = label
        self.on_press = on_press

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        if value is None:
            self._label = ''
        else:
            self._label = str(value)
        self._set_label(value)
        self.rehint()

    def _set_label(self, value):
        raise NotImplementedError('Button widget must define _set_label()')

    @property
    def on_press(self):
        return self._on_press

    @on_press.setter
    def on_press(self, handler):
        self._on_press = handler
        self._set_on_press(handler)

    def _set_on_press(self, value):
        pass


from .base import Widget


class Table(Widget):
    def __init__(self, headings, id=None, style=None):
        super().__init__(id=id, style=style)
        self.headings = headings

    def _configure(self):
        pass

    def insert(self, index, *data):
        if len(data) != len(self.headings):
            raise Exception('Data size does not match number of headings')

        if index is None:
            self._data.append(data)
        else:
            self._data.insert(index, data)



from colosseum import CSSNode


class Widget:
    '''This is the base widget implementation that all widgets in Toga
    derive from.

    It defines the interface for core functionality for children, styling,
    layout and ownership by specific App and Window.

    Apart from the above, this is an abstract implementation which must
    be made concrete by some platform-specific code for the _apply_layout
    method.

    :param id: An identifier for this widget.
    :param style: an optional style object. If no style is provided then a
                  new one will be created for the widget.
    '''
    def __init__(self, id=None, style=None, **config):
        self._id = id
        self._parent = None
        self._children = None
        self._window = None
        self._app = None
        self._impl = None
        self.__container = None
        self.dirty = True
        self._layout_in_progress = False
        self._config = config

        if style:
            self._style = style.apply(self)
        else:
            self._style = CSSNode(self)

    def __repr__(self):
        return "<%s:%s>" % (self.__class__.__name__, id(self))

    @property
    def id(self):
        '''The node identifier.

        This id can be used to target CSS directives
        '''
        return self._id

    @property
    def style(self):
        '''The style object for this widget.
        '''
        return self._style

    @property
    def parent(self):
        '''The parent of this node.
        '''
        return self._parent

    @property
    def children(self):
        '''The children of this node.

        This *always* returns a list, even if the node is a leaf
        and cannot have children.
        '''
        if self._children is None:
            return []
        else:
            return self._children

    def add(self, child):
        '''Add a widget as a child of this one.

        Raises an error if this widget is a leaf, and cannot
        have children.
        '''
        if self._children is None:
            raise ValueError('Widget cannot have children')

        self._children.append(child)

        child.app = self.app
        child._parent = self

        if self._parent:
            self._parent.dirty = True

        self._add_child(child)

    @property
    def app(self):
        '''The App to which this widget belongs.
        '''
        return self._app

    @app.setter
    def app(self, app):
        if self._app:
            raise Exception("Widget %r is already associated with an App" % self)
        self._app = app
        self._set_app(app)
        if self._children is not None:
            for child in self._children:
                child.app = app

    @property
    def window(self):
        '''The Window to which this widget belongs.
        '''
        return self._window

    @window.setter
    def window(self, window):
        self._window = window
        self._set_window(window)
        if self._children is not None:
            for child in self._children:
                child.window = window

    @property
    def _container(self):
        '''The display container to which this widget belongs.
        '''
        return self.__container

    @_container.setter
    def _container(self, container):
        self.__container = container
        self._set_container(container)
        if self._children is not None:
            for child in self._children:
                child._container = container

    def _create(self):
        self.create()
        self._configure(**self._config)

    def _initialize(self, **initial):
        pass

    def _update_layout(self, **style):
        """Force a layout update on the widget.

        The update request can be accompanied by additional style information
        (probably min_width, min_height, width or height) to control the
        layout.
        """
        # print("UPDATE LAYOUT OF ", self, style)
        if self._layout_in_progress:
            return
        self._layout_in_progress = True

        self.style.set(**style)

        # Recompute layout for this widget
        self.style.recompute()

        # Update the layout parameters for all children.
        # This will also perform a leaf-first update of
        # the constraint on each widget.
        self._update_child_layout()

        # Set the constraints the widget to adhere to the new style.
        self._apply_layout()

        self._layout_in_progress = False

    def _update_child_layout(self):
        # print("UPDATE CHILD LAYOUT - widget")
        if self._children is not None:
            for child in self.children:
                # if child.is_container:
                child._update_layout()



from .base import Widget


class ProgressBar(Widget):
    def __init__(self, id=None, style=None, max=None, value=None):
        super().__init__(id=id, style=style, max=max, value=value)

    def _configure(self, max, value):
        self.value = value
        self.max = max
        self.running = False
        self.rehint()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        self._running = self._value is not None

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, max):
        self._max = max
        self._set_max(max)



from .base import Widget


class OptionContainer(Widget):
    _CONTAINER_CLASS = None

    def __init__(self, id=None, style=None, content=None):
        super().__init__(id=id, style=style, content=content)
        self._containers = []

    def _configure(self, content):
        if content:
            for label, widget in content:
                self.add(label, widget)

    def add(self, label, widget):
        if widget._impl is None:
            container = self._CONTAINER_CLASS()
            container.content = widget
        else:
            container = widget

        self._containers.append((label, container, widget))

        widget.window = self.window
        widget.app = self.app

        self._add_content(label, container, widget)

    def _update_child_layout(self):
        for label, container, widget in self._containers:
            container._update_layout()



from .base import Widget


class Tree(Widget):
    def __init__(self, headings, id=None, style=None):
        super().__init__(id=id, style=style)
        self.headings = headings

    def _configure(self):
        pass
